@@ -1,6 +1,103 @@
 use crate::model::types::*;
 use crate::model::{Expression, Name};
-use std::mem::discriminant;
+use std::collections::HashMap;
+
+/// A builder pattern for `Module`s.
+pub struct ModuleBuilder {
+    module: Module,
+}
+
+impl ModuleBuilder {
+    /// Creates a new empty builder of WebAssembly modules.
+    pub fn new() -> Self {
+        ModuleBuilder {
+            module: Module::empty(),
+        }
+    }
+
+    /// Sets the function types segment for the WebAssembly module to be built.
+    pub fn set_function_types(&mut self, function_types: Vec<FunctionType>) {
+        self.module.function_types = Some(function_types);
+    }
+
+    /// Sets the functions segment for the WebAssembly module to be built.
+    pub fn set_functions(&mut self, functions: Vec<Function>) {
+        self.module.functions = Some(functions);
+    }
+
+    /// Sets the table segment for the WebAssembly module to be built.
+    pub fn set_tables(&mut self, tables: Vec<Table>) {
+        self.module.tables = Some(tables);
+    }
+
+    /// Sets the tables segment for the WebAssembly module to be built.
+    pub fn set_memories(&mut self, memories: Vec<Memory>) {
+        self.module.memories = Some(memories);
+    }
+
+    /// Sets the globals segment for the WebAssembly module to be built.
+    pub fn set_globals(&mut self, globals: Vec<Global>) {
+        self.module.globals = Some(globals);
+    }
+
+    /// Sets the elements segment for the WebAssembly module to be built.
+    pub fn set_elements(&mut self, elements: Vec<Element>) {
+        self.module.elements = Some(elements);
+    }
+
+    /// Sets the data segment for the WebAssembly module to be built.
+    pub fn set_data(&mut self, data: Vec<Data>) {
+        self.module.data = Some(data);
+    }
+
+    /// Sets the start segment for the WebAssembly module to be built.
+    pub fn set_start(&mut self, start: Start) {
+        self.module.start = Some(start);
+    }
+
+    /// Sets the imports segment for the WebAssembly module to be built.
+    pub fn set_imports(&mut self, imports: Vec<Import>) {
+        self.module.imports = Some(imports);
+    }
+
+    /// Sets the exports segment for the WebAssembly module to be built.
+    pub fn set_exports(&mut self, exports: Vec<Export>) {
+        self.module.exports = Some(exports);
+    }
+
+    /// Sets the custom section at the given insertion point for the WebAssembly module to be built.
+    pub fn set_custom_sections(
+        &mut self,
+        insertion_point: ModuleSection,
+        custom_sections: Vec<Custom>,
+    ) {
+        self.module.customs.insert(insertion_point, custom_sections);
+    }
+
+    /// Determines whether the WebAssembly module to be built will include a data count section or not.  
+    pub fn include_data_count(&mut self, include: bool) {
+        self.module.data_count = include;
+    }
+
+    /// Builds the current segments into a module.
+    pub fn build(self) -> Module {
+        self.into()
+    }
+}
+
+impl From<ModuleBuilder> for Module {
+    fn from(builder: ModuleBuilder) -> Self {
+        builder.module
+    }
+}
+
+impl Default for ModuleBuilder {
+    fn default() -> Self {
+        ModuleBuilder {
+            module: Module::empty(),
+        }
+    }
+}
 
 /// WebAssembly programs are organized into modules, which are the unit of deployment, loading, and compilation.
 /// A module collects definitions for types, functions, tables, memories, and globals.
@@ -10,119 +107,122 @@ use std::mem::discriminant;
 /// Each of the vectors – and thus the entire module – may be empty.
 ///
 /// See https://webassembly.github.io/spec/core/syntax/modules.html#modules
-#[derive(Clone, Debug, PartialEq)]
+///
+/// # Examples
+/// ## Empty
+/// ```rust
+/// use wasm_ast::{Module, ModuleSection};
+/// use std::collections::HashMap;
+///
+/// let module = Module::empty();
+///
+/// assert_eq!(module.functions(), None);
+/// assert_eq!(module.functions(), None);
+/// assert_eq!(module.tables(), None);
+/// assert_eq!(module.memories(), None);
+/// assert_eq!(module.globals(), None);
+/// assert_eq!(module.elements(), None);
+/// assert_eq!(module.data(), None);
+/// assert_eq!(module.start(), None);
+/// assert_eq!(module.imports(), None);
+/// assert_eq!(module.exports(), None);
+/// assert_eq!(module.include_data_count(), false);
+/// ```
+///
+/// ## Builder
+/// ```rust
+/// use wasm_ast::{Module, Import, FunctionType, ValueType, Start, Function, ResultType, ControlInstruction, Memory, Limit, Export, Data, Expression, ModuleSection, Custom};
+/// use std::collections::HashMap;
+///
+/// let mut module = Module::builder();
+/// let module = module.build();
+///
+/// assert_eq!(module.functions(), None);
+/// assert_eq!(module.functions(), None);
+/// assert_eq!(module.tables(), None);
+/// assert_eq!(module.memories(), None);
+/// assert_eq!(module.globals(), None);
+/// assert_eq!(module.elements(), None);
+/// assert_eq!(module.data(), None);
+/// assert_eq!(module.start(), None);
+/// assert_eq!(module.imports(), None);
+/// assert_eq!(module.exports(), None);
+/// assert_eq!(module.include_data_count(), false);
+/// ```
+#[derive(Clone, Debug)]
 pub struct Module {
-    function_types: Vec<FunctionType>,
-    functions: Vec<Function>,
-    tables: Vec<Table>,
-    memories: Vec<Memory>,
-    globals: Vec<Global>,
-    elements: Vec<Element>,
-    data: Vec<Data>,
+    function_types: Option<Vec<FunctionType>>,
+    functions: Option<Vec<Function>>,
+    tables: Option<Vec<Table>>,
+    memories: Option<Vec<Memory>>,
+    globals: Option<Vec<Global>>,
+    elements: Option<Vec<Element>>,
+    data: Option<Vec<Data>>,
     start: Option<Start>,
-    imports: Vec<Import>,
-    exports: Vec<Export>,
-    customs: Vec<Custom>,
+    imports: Option<Vec<Import>>,
+    exports: Option<Vec<Export>>,
+    customs: HashMap<ModuleSection, Vec<Custom>>,
+    data_count: bool,
 }
 
 impl Module {
+    /// Creates a builder for WebAssembly modules.
+    pub fn builder() -> ModuleBuilder {
+        ModuleBuilder::new()
+    }
+
     /// Creates a new empty `Module`.
-    pub fn new() -> Self {
+    pub fn empty() -> Self {
         Module {
-            function_types: vec![],
-            functions: vec![],
-            tables: vec![],
-            memories: vec![],
-            globals: vec![],
-            elements: vec![],
-            data: vec![],
+            function_types: None,
+            functions: None,
+            tables: None,
+            memories: None,
+            globals: None,
+            elements: None,
+            data: None,
             start: None,
-            imports: vec![],
-            exports: vec![],
-            customs: vec![],
+            imports: None,
+            exports: None,
+            customs: HashMap::new(),
+            data_count: false,
         }
     }
 
     /// The 𝗍𝗒𝗉𝖾𝗌 component of a module defines a vector of function types.
-    pub fn types(&self) -> &[FunctionType] {
-        &self.function_types
-    }
-
-    pub fn add_type(&mut self, function_type: FunctionType) -> TypeIndex {
-        self.function_types.push(function_type);
-        self.function_types.len() - 1
+    pub fn function_types(&self) -> Option<&[FunctionType]> {
+        self.function_types.as_deref()
     }
 
     /// The 𝖿𝗎𝗇𝖼𝗌 component of a module defines a vector of functions.
-    pub fn functions(&self) -> &[Function] {
-        &self.functions
-    }
-
-    pub fn add_function(&mut self, function: Function) {
-        self.functions.push(function);
+    pub fn functions(&self) -> Option<&[Function]> {
+        self.functions.as_deref()
     }
 
     /// The 𝗍𝖺𝖻𝗅𝖾𝗌 component of a module defines a vector of tables described by their table type.
-    pub fn tables(&self) -> &[Table] {
-        &self.tables
-    }
-
-    pub fn add_table(&mut self, table: Table) -> TableIndex {
-        self.tables.push(table);
-
-        let imported_tables = self
-            .imports()
-            .iter()
-            .filter(|import| matches!(import.description(), ImportDescription::Table(_)))
-            .count();
-
-        imported_tables + self.tables.len() - 1
+    pub fn tables(&self) -> Option<&[Table]> {
+        self.tables.as_deref()
     }
 
     /// The 𝗆𝖾𝗆𝗌 component of a module defines a vector of linear memories (or memories for short)
     /// as described by their memory type.
-    pub fn memories(&self) -> &[Memory] {
-        &self.memories
-    }
-
-    pub fn add_memory(&mut self, memory: Memory) -> MemoryIndex {
-        self.memories.push(memory);
-
-        let imported_memories = self
-            .imports()
-            .iter()
-            .filter(|import| matches!(import.description(), ImportDescription::Memory(_)))
-            .count();
-
-        imported_memories + self.memories.len() - 1
+    pub fn memories(&self) -> Option<&[Memory]> {
+        self.memories.as_deref()
     }
 
     /// The 𝗀𝗅𝗈𝖻𝖺𝗅𝗌 component of a module defines a vector of global variables (or globals for short).
-    pub fn globals(&self) -> &[Global] {
-        &self.globals
-    }
-
-    pub fn add_global(&mut self, global: Global) {
-        self.globals.push(global);
+    pub fn globals(&self) -> Option<&[Global]> {
+        self.globals.as_deref()
     }
 
     /// The 𝖾𝗅𝖾𝗆𝗌 component of a module defines a vector of element segments.
-    pub fn elements(&self) -> &[Element] {
-        &self.elements
-    }
-
-    pub fn add_element(&mut self, element: Element) -> ElementIndex {
-        self.elements.push(element);
-        self.elements().len() - 1
+    pub fn elements(&self) -> Option<&[Element]> {
+        self.elements.as_deref()
     }
 
     /// The 𝖽𝖺𝗍𝖺𝗌 component of a module defines a vector of data segments.
-    pub fn data(&self) -> &[Data] {
-        &self.data
-    }
-
-    pub fn add_data(&mut self, data: Data) {
-        self.data.push(data);
+    pub fn data(&self) -> Option<&[Data]> {
+        self.data.as_deref()
     }
 
     /// The 𝗌𝗍𝖺𝗋𝗍 component of a module declares the function index of a start function that is
@@ -131,41 +231,26 @@ impl Module {
         self.start.as_ref()
     }
 
-    pub fn set_start(&mut self, start: Option<Start>) {
-        self.start = start;
-    }
-
     /// The 𝗂𝗆𝗉𝗈𝗋𝗍𝗌 component of a module defines a set of imports that are required for instantiation.
-    pub fn imports(&self) -> &[Import] {
-        &self.imports
-    }
-
-    pub fn add_import(&mut self, import: Import) -> usize {
-        let import_discriminant = discriminant(import.description());
-
-        self.imports.push(import);
-
-        self.imports()
-            .iter()
-            .filter(|i| discriminant(i.description()) == import_discriminant)
-            .count()
-            - 1
+    pub fn imports(&self) -> Option<&[Import]> {
+        self.imports.as_deref()
     }
 
     /// The 𝖾𝗑𝗉𝗈𝗋𝗍𝗌 component of a module defines a set of exports that become accessible to the
     /// host environment once the module has been instantiated.
-    pub fn exports(&self) -> &[Export] {
-        &self.exports
+    pub fn exports(&self) -> Option<&[Export]> {
+        self.exports.as_deref()
     }
 
-    pub fn add_export(&mut self, export: Export) {
-        self.exports.push(export);
+    /// The custom sections of a module for a given insertion point.
+    /// Custom sections are allowed at the beginning of a module and after every other section.
+    pub fn custom_sections_at(&self, insertion_point: ModuleSection) -> Option<&[Custom]> {
+        self.customs.get(&insertion_point).map(Vec::as_slice)
     }
-}
 
-impl Default for Module {
-    fn default() -> Self {
-        Self::new()
+    /// Whether the module includes the data count section or not.
+    pub fn include_data_count(&self) -> bool {
+        self.data_count
     }
 }
 
@@ -1032,109 +1117,59 @@ pub enum ImportDescription {
     Global(GlobalType),
 }
 
-#[cfg(test)]
-mod tests {
-    use super::*;
-    use crate::model::{ControlInstruction, Instruction, NumericInstruction};
-
-    #[test]
-    fn empty_module() {
-        let module = Module::new();
-
-        assert!(module.types().is_empty());
-        assert!(module.functions().is_empty());
-        assert!(module.tables().is_empty());
-        assert!(module.memories().is_empty());
-        assert!(module.globals().is_empty());
-        assert!(module.imports().is_empty());
-        assert!(module.exports().is_empty());
-        assert!(module.data().is_empty());
-        assert!(module.elements().is_empty());
-        assert!(module.start().is_none());
-    }
-
-    #[test]
-    fn module() {
-        let mut module = Module::new();
-        let function_type = FunctionType::new(
-            ResultType::from(vec![IntegerType::I64.into()]),
-            ResultType::from(vec![FloatType::F64.into()]),
-        );
-        module.add_type(function_type.clone());
-
-        let function = Function::new(
-            0,
-            ResultType::from(vec![IntegerType::I32.into()]),
-            Expression::new(vec![Instruction::Control(ControlInstruction::Nop)]),
-        );
-        module.add_function(function.clone());
-
-        let element = Element::new(
-            ReferenceType::Function,
-            ElementMode::Passive,
-            ElementInitializer::FunctionIndex(vec![0]),
-        );
-        module.add_element(element.clone());
-
-        let data = Data::new(DataMode::Passive, vec![42]);
-        module.add_data(data.clone());
-
-        let table = Table::new(TableType::new(Limit::new(0, None), ReferenceType::Function));
-        module.add_table(table);
-
-        let memory = Memory::new(MemoryType::from(Limit::new(0, None)));
-        module.add_memory(memory);
-
-        let import = Import::new(
-            "test".into(),
-            "foobar".into(),
-            ImportDescription::Function(0),
-        );
-        module.add_import(import.clone());
-
-        let export = Export::new("foobar".into(), ExportDescription::Function(0));
-        module.add_export(export.clone());
-
-        let start = Start::new(0);
-        module.set_start(Some(start));
-
-        let global = Global::new(
-            GlobalType::immutable(IntegerType::I64.into()),
-            Expression::new(vec![Instruction::Numeric(NumericInstruction::I64Constant(
-                0,
-            ))]),
-        );
-        module.add_global(global.clone());
-
-        assert_eq!(module.types(), &[function_type]);
-        assert_eq!(module.functions(), &[function]);
-        assert_eq!(module.tables(), &[table]);
-        assert_eq!(module.memories(), &[memory]);
-        assert_eq!(module.globals(), &[global]);
-        assert_eq!(module.imports(), &[import]);
-        assert_eq!(module.exports(), &[export]);
-        assert_eq!(module.data(), &[data]);
-        assert_eq!(module.elements(), &[element]);
-        assert_eq!(module.start(), Some(&start));
-    }
-
-    #[test]
-    fn new_function() {
-        let kind = 1;
-        let locals = ResultType::from(vec![IntegerType::I64.into()]);
-        let body = Expression::new(Vec::new());
-        let function = Function::new(kind, locals.clone(), body.clone());
-
-        assert_eq!(function.kind(), 1);
-        assert_eq!(function.locals(), &locals);
-        assert_eq!(function.body(), &body);
-    }
-
-    #[test]
-    fn new_table() {
-        let kind = TableType::new(Limit::new(0, None), ReferenceType::Function);
-        let table = Table::new(kind);
-
-        assert_eq!(table.kind(), &kind);
-    }
+/// The binary encoding of modules is organized into sections.
+/// Most sections correspond to one component of a module record,
+/// except that function definitions are split into two sections,
+/// separating their type declarations in the function section from their bodies in the code section.
+///
+/// See https://webassembly.github.io/spec/core/binary/modules.html
+#[derive(Copy, Clone, Debug, Eq, PartialEq, Ord, PartialOrd, Hash)]
+pub enum ModuleSection {
+    /// Custom sections have the id 0.
+    /// They are intended to be used for debugging information or third-party extensions,
+    /// and are ignored by the WebAssembly semantics.
+    /// Their contents consist of a name further identifying the custom section,
+    /// followed by an uninterpreted sequence of bytes for custom use.
+    Custom = 0,
+    /// The type section has the id 1.
+    /// It decodes into a vector of function types that represent the 𝗍𝗒𝗉𝖾𝗌 component of a module.
+    Type,
+    /// The import section has the id 2.
+    /// It decodes into a vector of imports that represent the 𝗂𝗆𝗉𝗈𝗋𝗍𝗌 component of a module.
+    Import,
+    /// The function section has the id 3.
+    /// It decodes into a vector of type indices that represent the 𝗍𝗒𝗉𝖾 fields of the functions
+    /// in the 𝖿𝗎𝗇𝖼𝗌 component of a module. The 𝗅𝗈𝖼𝖺𝗅𝗌 and 𝖻𝗈𝖽𝗒 fields of the respective functions
+    /// are encoded separately in the code section.
+    Function,
+    /// The table section has the id 4.
+    /// It decodes into a vector of tables that represent the 𝗍𝖺𝖻𝗅𝖾𝗌 component of a module.
+    Table,
+    /// The memory section has the id 5.
+    /// It decodes into a vector of memories that represent the 𝗆𝖾𝗆𝗌 component of a module.
+    Memory,
+    /// The global section has the id 6.
+    /// It decodes into a vector of globals that represent the 𝗀𝗅𝗈𝖻𝖺𝗅𝗌 component of a module.
+    Global,
+    /// The export section has the id 7.
+    /// It decodes into a vector of exports that represent the 𝖾𝗑𝗉𝗈𝗋𝗍𝗌 component of a module.
+    Export,
+    /// The start section has the id 8.
+    /// It decodes into an optional start function that represents the 𝗌𝗍𝖺𝗋𝗍 component of a module.
+    Start,
+    /// The element section has the id 9.
+    /// It decodes into a vector of element segments that represent the 𝖾𝗅𝖾𝗆𝗌 component of a module.
+    Element,
+    /// The code section has the id 10.
+    /// It decodes into a vector of code entries that are pairs of value type vectors and expressions.
+    /// They represent the 𝗅𝗈𝖼𝖺𝗅𝗌 and 𝖻𝗈𝖽𝗒 field of the functions in the 𝖿𝗎𝗇𝖼𝗌 component of a module.
+    /// The 𝗍𝗒𝗉𝖾 fields of the respective functions are encoded separately in the function section.
+    Code,
+    /// The data section has the id 11.
+    /// It decodes into a vector of data segments that represent the 𝖽𝖺𝗍𝖺𝗌 component of a module.
+    Data,
+    /// The data count section has the id 12.
+    /// It decodes into an optional u32 that represents the number of data segments in the data section.
+    /// If this count does not match the length of the data segment vector, the module is malformed.
+    DataCount,
 }
